#include <sys/queue.h>

#include "msph/tree.h"
#include "spho/scope.h"
#include "spho/spho.h"
#include "spho/tp.h"

#include "msph/sphophi.h"

typedef int (*msph_tree_sphophi_f)(struct spho_scope *, struct msph_tree *);

static int	msph_tree_sphophi_decor_scope(struct spho_scope *,
		    struct msph_tree *);
static int	msph_tree_sphophi_decor_tpexpr_nom(struct spho_scope *,
		    struct msph_tree *);
static int	msph_tree_sphophi_decor_tp(struct spho_scope *,
		    struct msph_tree *);
static int	msph_tree_sphophi_tp_prebind(struct spho_scope *,
		    struct msph_tree *);

/* is the tree a name binding? */
#define IS_BINDING(tree)	(TREE_ID(tree) == MSPH_TREE_TPDEF	|| \
				 TREE_ID(tree) == MSPH_TREE_NOMINDECL	|| \
				 TREE_ID(tree) == MSPH_TREE_MEMBDECL	|| \
				 TREE_ID(tree) == MSPH_TREE_FORALL)
/* is the tree a type expression? */
#define IS_TPEXPR(tree)		((tree)->type & MSPH_TREE_TPEXPR)

/* foreach_pre: top down foreach */
static int	msph_sphophi_foreach_pre(msph_tree_sphophi_f,
		    struct msph_tree_root *);
static int	msph_sphophi_foreach_pre_unit(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_unit *);
static int	msph_sphophi_foreach_pre_body(msph_tree_sphophi_f, 
		    struct spho_scope *, struct msph_tree_body *);
static int	msph_sphophi_foreach_pre_dir(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_dir *);
static int	msph_sphophi_foreach_pre_assert(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_assert *);
static int	msph_sphophi_foreach_pre_membdecl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_membdecl *);
static int	msph_sphophi_foreach_pre_namedecl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_namedecl *);
static int	msph_sphophi_foreach_pre_nameref(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_nameref *);
static int	msph_sphophi_foreach_pre_tpdef(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_tpdef *);
static int	msph_sphophi_foreach_pre_nomindecl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_nomindecl *);
static int	msph_sphophi_foreach_pre_tpexpr(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_tpexpr *);
static int	msph_sphophi_foreach_pre_true(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_true *);
static int	msph_sphophi_foreach_pre_false(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_false *);
static int	msph_sphophi_foreach_pre_tpname(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_tpname *);
static int	msph_sphophi_foreach_pre_tpappl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_tpappl *);
static int	msph_sphophi_foreach_pre_trait(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_trait *);
static int	msph_sphophi_foreach_pre_conj(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_conj *);
static int	msph_sphophi_foreach_pre_disj(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_disj *);
static int	msph_sphophi_foreach_pre_impl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_impl *);
static int	msph_sphophi_foreach_pre_arrow(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_arrow *);
static int	msph_sphophi_foreach_pre_box(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_box *);
static int	msph_sphophi_foreach_pre_forall(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_forall *);
static int	msph_sphophi_foreach_pre_paren(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_paren *);

/* foreach_post: bottom up foreach */
static int	msph_sphophi_foreach_post(msph_tree_sphophi_f,
		    struct msph_tree_root *);
static int	msph_sphophi_foreach_post_unit(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_unit *);
static int	msph_sphophi_foreach_post_body(msph_tree_sphophi_f, 
		    struct spho_scope *, struct msph_tree_body *);
static int	msph_sphophi_foreach_post_dir(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_dir *);
static int	msph_sphophi_foreach_post_assert(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_assert *);
static int	msph_sphophi_foreach_post_membdecl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_membdecl *);
static int	msph_sphophi_foreach_post_tpdef(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_tpdef *);
static int	msph_sphophi_foreach_post_nomindecl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_nomindecl *);
static int	msph_sphophi_foreach_post_tpexpr(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_tpexpr *);
static int	msph_sphophi_foreach_post_true(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_true *);
static int	msph_sphophi_foreach_post_false(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_false *);
static int	msph_sphophi_foreach_post_tpname(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_tpname *);
static int	msph_sphophi_foreach_post_tpappl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_tpappl *);
static int	msph_sphophi_foreach_post_trait(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_trait *);
static int	msph_sphophi_foreach_post_conj(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_conj *);
static int	msph_sphophi_foreach_post_disj(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_disj *);
static int	msph_sphophi_foreach_post_impl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_impl *);
static int	msph_sphophi_foreach_post_arrow(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_arrow *);
static int	msph_sphophi_foreach_post_box(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_box *);
static int	msph_sphophi_foreach_post_forall(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_forall *);
static int	msph_sphophi_foreach_post_paren(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_paren *);
static int	msph_sphophi_foreach_post_namedecl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_namedecl *);
static int	msph_sphophi_foreach_post_nameref(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_nameref *);

int msph_sphophi_init(struct msph_tree_root *root)
{
	if ((root->ctx->spho = spho_ctx_create()) == NULL) {
		MSPH_ERR(root->ctx, MSPH_ERR_SYS);
		return (-1);
	}

	return (0);
}

int
msph_sphophi_scope(struct msph_tree_root *root)
{
	return (msph_sphophi_foreach_pre(msph_tree_sphophi_decor_scope, root));
}

int
msph_sphophi_nom_lookup(struct msph_tree_root *root)
{
	return (msph_sphophi_foreach_pre(msph_tree_sphophi_decor_tpexpr_nom,
		    root));
}

int
msph_sphophi_tp(struct msph_tree_root *root)
{
	return (msph_sphophi_foreach_post(msph_tree_sphophi_decor_tp, root));
}

int
msph_sphophi_tp_prebind(struct msph_tree_root *root)
{
	return (msph_sphophi_foreach_pre(msph_tree_sphophi_tp_prebind,
		    root));
}

int msph_sphophi(struct msph_tree_root *root)
{
	int ret;

	if ((ret = msph_sphophi_init(root)) != 0)
		return (ret);
	if ((ret = msph_sphophi_scope(root)) != 0)
		return (ret);
	if ((ret = msph_sphophi_nom_lookup(root)) != 0)
		return (ret);
	if ((ret = msph_sphophi_tp(root)) != 0)
		return (ret);
	if ((ret = msph_sphophi_tp_prebind(root)) != 0)
		return (ret);

	return (ret);
}



static int
msph_tree_sphophi_decor_scope(struct spho_scope *sc, struct msph_tree *tree)
{
	int ret;
	struct msph_tree_ident *ident;
	struct msph_tree_tpdef *tpdef;

	ret = 0;

	switch (TREE_ID(tree)) {
	case MSPH_TREE_UNIT:
	case MSPH_TREE_TRAIT:
	case MSPH_TREE_FORALL:
		if ((SCOPE(tree) = spho_scope_create(sc)) == NULL)
			ret = -1;
		ADD_FLAG(tree, MSPH_TREE_FLAG_SCOPE);
		break;
	case MSPH_TREE_TPDEF:
		tpdef = (struct msph_tree_tpdef *)tree;
		if (! STAILQ_EMPTY(&tpdef->params) &&
		    ((SCOPE(tpdef) = spho_scope_create(sc)) == NULL)) {
			ret = -1;
		}
		ADD_FLAG(tree, MSPH_TREE_FLAG_SCOPE);
		break;
	case MSPH_TREE_IDENT:
		if (! IS_BINDING(tree->parent))
			break;
		ident = (struct msph_tree_ident *)tree;
		if ((NOM(ident) = spho_scope_nom_add(sc, ident->str,
			    sizeof(ident->str))) == NULL)
			ret = -1;
		ADD_FLAG(tree, MSPH_TREE_FLAG_NOM);
		break;
	default:
		break;
	}

	return (ret);
}

static int
msph_tree_sphophi_decor_tpexpr_nom(struct spho_scope *sc,
    struct msph_tree *tree)
{
	int ret;
	struct msph_tree_ident *ident;

	ret = 0;

	switch (TREE_ID(tree)) {
	case MSPH_TREE_IDENT:
		if (! IS_TPEXPR(tree->parent))
			break;
		ident = (struct msph_tree_ident *)tree;
		ret = spho_scope_nom_lookup_str_strict(sc, ident->str,
			    sizeof(ident->str), &NOM(ident));

		break;
	default:
		break;
	}

	return (ret);
}

static int
msph_tree_sphophi_decor_tp(struct spho_scope *sc, struct msph_tree *tree)
{
	int ret;
	struct msph_tree_tpexpr *tpexpr_i;
	struct msph_tree_conj *conj;
	struct msph_tree_disj *disj;
	struct msph_tree_impl *impl;
	struct msph_tree_arrow *arrow;
	struct msph_tree_tpname *tpname;
	struct msph_tree_tpappl *tpappl;
	struct msph_tree_box *box;
	struct msph_tree_forall *forall;
	struct msph_tree_paren *paren;
	struct msph_tree_trait *trait;
	struct msph_tree_dir *dir;
	struct msph_tree_membdecl *membd;

	struct spho_tp_l *tps;
	struct spho_tp *tp, *tp_memb;

	ret = -1;
	tps = NULL;
	tp = tp_memb = NULL;
	dir = NULL;
	membd = NULL;

	if (! (tree->type & MSPH_TREE_TPEXPR))
		return (0);

	switch (TREE_ID(tree)) {
	case MSPH_TREE_CONJ:
		conj = (struct msph_tree_conj *)tree;
		if ((TP(conj) = spho_tp_create_conj(sc, TP(conj->ltp),
		    TP(conj->rtp))) == NULL)
			goto cleanup;
		ADD_FLAG(conj, MSPH_TREE_FLAG_TP);
		break;
	case MSPH_TREE_DISJ:
		disj = (struct msph_tree_disj *)tree;
		if ((TP(disj) = spho_tp_create_conj(sc, TP(disj->ltp),
		    TP(disj->rtp))) == NULL)
			goto cleanup;
		ADD_FLAG(disj, MSPH_TREE_FLAG_TP);
		break;
	case MSPH_TREE_IMPL:
		impl = (struct msph_tree_impl *)tree;
		if ((TP(impl) = spho_tp_create_conj(sc, TP(impl->ltp),
		    TP(impl->rtp))) == NULL)
			goto cleanup;
		ADD_FLAG(impl, MSPH_TREE_FLAG_TP);
		break;
	case MSPH_TREE_ARROW:
		arrow = (struct msph_tree_arrow *)tree;
		if ((TP(arrow) = spho_tp_create_conj(sc, TP(arrow->ltp),
		    TP(arrow->rtp))) == NULL)
			goto cleanup;
		ADD_FLAG(arrow, MSPH_TREE_FLAG_TP);
		break;
	case MSPH_TREE_TPNAME:
		tpname = (struct msph_tree_tpname *)tree;
		if ((TP(tpname) = spho_tp_create_name(sc, NOM(tpname->name)))
		    == NULL)
			goto cleanup;
		ADD_FLAG(tpname, MSPH_TREE_FLAG_TP);
		break;
	case MSPH_TREE_TPAPPL:
		tpappl = (struct msph_tree_tpappl *)tree;
		if ((tps = malloc(sizeof(*tps))) == NULL) {
			SPHO_SC_ERR(sc, SPHO_ERR_SYS);
			goto cleanup;
		}
		STAILQ_INIT(tps);
		STAILQ_FOREACH(tpexpr_i, &tpappl->args, entries) {
			STAILQ_INSERT_TAIL(tps, TP(tpexpr_i), entries);
		}
		if ((TP(tpappl) = spho_tp_create_appl(sc, NOM(tpappl->name),
			    tps)) == NULL)
			goto cleanup;
		tps = NULL;
		ADD_FLAG(tpappl, MSPH_TREE_FLAG_TP);
		break;
	case MSPH_TREE_TRUE:
		if ((TP(tree) = spho_tp_create_true(sc)) == NULL)
			goto cleanup;
		ADD_FLAG(tree, MSPH_TREE_FLAG_TP);
		break;
	case MSPH_TREE_FALSE:
		if ((TP(tree) = spho_tp_create_false(sc)) == NULL)
			goto cleanup;
		ADD_FLAG(tree, MSPH_TREE_FLAG_TP);
		break;
	case MSPH_TREE_FORALL:
		forall = (struct msph_tree_forall *)tree;
		if ((TP(forall) = spho_tp_create_forall(sc,
			    NOM(forall->name), TP(forall->inner))) == NULL)
			goto cleanup;
		ADD_FLAG(forall, MSPH_TREE_FLAG_TP);
		break;
	case MSPH_TREE_BOX:
		box = (struct msph_tree_box *)tree;
		if ((TP(box) = spho_tp_create_box(sc, TP(box->inner)))
		    == NULL)
			goto cleanup;
		ADD_FLAG(box, MSPH_TREE_FLAG_TP);
		break;
	case MSPH_TREE_PAREN:
		paren = (struct msph_tree_paren *)tree;
		if ((TP(paren) = TP(paren->inner)) == NULL)
			goto cleanup;
		ADD_FLAG(paren, MSPH_TREE_FLAG_TP);
		break;
	case MSPH_TREE_TRAIT:
		trait = (struct msph_tree_trait *)tree;

		STAILQ_FOREACH(dir, &trait->body->head, entries) {
			if (TREE_ID(dir) != MSPH_TREE_MEMBDECL)
				continue;
			membd = (struct msph_tree_membdecl *)dir;
			// TODO check that we make sure to dealloc when
			// spho has stabilized a bit
			if ((tp_memb = spho_tp_create_member(sc, NOM(membd->name),
			    TP(membd->tp))) == NULL)
				goto cleanup;
			if (tp == NULL) {
				tp = tp_memb;
			} else {
				if ((tp = spho_tp_create_conj(sc, tp, tp_memb))
				    == NULL)
					goto cleanup;
			}
		}
		if (tp == NULL)
			tp = spho_tp_create_true(sc);
		SPHO_ASSERT(tp != NULL);
		TP(trait) = tp;
		ADD_FLAG(trait, MSPH_TREE_FLAG_TP);
		break;
	default:
		SPHO_ASSERT(0);
		break;
	}

	ret = 0;

cleanup:
	if (tps != NULL)
		free(tps);

	return (ret);
}

static int
msph_tree_sphophi_tp_prebind(struct spho_scope *sc,
    struct msph_tree *tree)
{
	int ret;
	struct spho_tp_op *op;
	struct msph_tree_tpdef *tpdef;
	struct msph_tree_namedecl *name;

	ret = 0;

	if (HAS_SCOPE(tree)) {
		if (spho_scope_prebind_init(SCOPE(tree)) == -1)
			return (-1);
	}

	switch (TREE_ID(tree)) {
	case MSPH_TREE_TPDEF:
		tpdef = (struct msph_tree_tpdef *)tree;

		if (HAS_PARAMS(tpdef)) {
			SPHO_ASSERT(HAS_SCOPE(tpdef));
			SPHO_ASSERT(SCOPE(tpdef) != NULL);
			SPHO_ASSERT(SCOPE(tpdef)->bind_loc != NULL);

			STAILQ_FOREACH(name, &tpdef->params, entries) {
				if (spho_scope_prebind_undef(SCOPE(tpdef),
					    NOM(name)) == -1) {
					return (-1);
				}
			}

			if ((op = spho_tp_op_create(sc, SCOPE(tpdef),
				    TP(tpdef->tp))) == NULL) {
				return (-1);
			}

			if ((ret = spho_scope_prebind_tp_op(sc, NOM(name), op))
			    == -1) {
				return (-1);
			}
		} else {
			if ((ret = spho_scope_prebind_tp(sc, NOM(tpdef->name),
				    TP(tpdef->tp))) == -1) {
				return (-1);
			}
		}
		break;
	default:
		break;
	}

	return (ret);
}

/* foreach pre: top down foreach node */
static int
msph_sphophi_foreach_pre(msph_tree_sphophi_f f, struct msph_tree_root *root)
{
	int ret;
	struct spho_scope *sc;
	struct msph_tree_unit *unit;

	SPHO_PRECOND(f != NULL);
	SPHO_PRECOND(root != NULL);
	SPHO_PRECOND(root->ctx != NULL);
	SPHO_PRECOND(root->ctx->spho != NULL);

	sc = spho_scope_global(root->ctx->spho);

	if ((ret = f(sc, TREE(root))) != 0) {
		goto err;
	}

	ret = 0;
	STAILQ_FOREACH(unit, &root->head, entries) {
		if ((ret = msph_sphophi_foreach_pre_unit(f, sc, unit)) != 0)
			goto err;
	}

	return (ret);
err:
	MSPH_ERR(root->ctx, MSPH_ERR_SPHOPHI);
	return (ret);
}

static int
msph_sphophi_foreach_pre_unit(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_unit *unit)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(unit != NULL);

	if ((ret = f(sc, TREE(unit))) != 0)
		return (ret);

	sc = SCOPE(unit);

	return (msph_sphophi_foreach_pre_body(f, sc, unit->body));
}

static int
msph_sphophi_foreach_pre_body(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_body *body)
{
	int ret;
	struct msph_tree_dir *dir;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(body != NULL);

	if ((ret = f(sc, TREE(body))) != 0)
		return (ret);

	ret = 0;
	STAILQ_FOREACH(dir, &body->head, entries) {
		if ((ret = msph_sphophi_foreach_pre_dir(f, sc, dir)) != 0)
			return (ret);
	}

	return (ret);
}

static int
msph_sphophi_foreach_pre_dir(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_dir *dir)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(dir != NULL);

	ret = -1;
	switch (TREE_ID(dir)) {
	case MSPH_TREE_TPDEF:
		ret = msph_sphophi_foreach_pre_tpdef(f, sc,
		    (struct msph_tree_tpdef *)dir);
		break;
	case MSPH_TREE_MEMBDECL:
		ret = msph_sphophi_foreach_pre_membdecl(f, sc,
		    (struct msph_tree_membdecl *)dir);
		break;
	case MSPH_TREE_NOMINDECL:
		ret = msph_sphophi_foreach_pre_nomindecl(f, sc,
		    (struct msph_tree_nomindecl *)dir);
		break;
	case MSPH_TREE_ASSERT:
		ret = msph_sphophi_foreach_pre_assert(f, sc,
		    (struct msph_tree_assert *)dir);
		break;
	default:
		SPHO_ASSERT(0);
		break;
	}

	return (ret);
}

static int
msph_sphophi_foreach_pre_tpdef(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_tpdef *tpdef)
{
	int ret;
	struct msph_tree_namedecl *name;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(tpdef != NULL);

	if ((ret = f(sc, TREE(tpdef))) != 0)
		return (ret);

	if ((ret = msph_sphophi_foreach_pre_namedecl(f, sc, tpdef->name)) != 0)
		return (ret);

	sc = HAS_SCOPE(tpdef) ? SCOPE(tpdef) : sc;

	STAILQ_FOREACH(name, &tpdef->params, entries) {
		if ((ret = msph_sphophi_foreach_pre_namedecl(f, sc, name))
		    != 0) {
			return (ret);
		}
	}

	return (msph_sphophi_foreach_pre_tpexpr(f, sc, tpdef->tp));
}

static int
msph_sphophi_foreach_pre_nomindecl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_nomindecl *nomd)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(nomd != NULL);

	if ((ret = f(sc, TREE(nomd))) != 0)
		return (ret);

	return (msph_sphophi_foreach_pre_namedecl(f, sc, nomd->name));
}

static int
msph_sphophi_foreach_pre_membdecl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_membdecl *membd)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(membd != NULL);

	if ((ret = f(sc, TREE(membd))) != 0)
		return (ret);

	if (msph_sphophi_foreach_pre_namedecl(f, sc, membd->name) != 0)
		return (ret);
	if (msph_sphophi_foreach_pre_tpexpr(f, sc, membd->tp) != 0)
		return (ret);

	return (0);
}

static int
msph_sphophi_foreach_pre_assert(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_assert *ass)
{
	int ret;

	if ((ret = f(sc, TREE(ass))) != 0)
		return (ret);

	if ((ret = msph_sphophi_foreach_pre_tpexpr(f, sc, ass->ltp)) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_pre_tpexpr(f, sc, ass->rtp)) != 0)
		return (ret);

	return (ret);
}

static int
msph_sphophi_foreach_pre_namedecl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_namedecl *name)
{
	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(name != NULL);

	return (f(sc, TREE(name)));
}

static int
msph_sphophi_foreach_pre_nameref(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_nameref *name)
{
	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(name != NULL);

	return (f(sc, TREE(name)));
}

static int
msph_sphophi_foreach_pre_tpexpr(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_tpexpr *tpexpr)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(tpexpr != NULL);

	ret = -1;

	switch (TREE_ID(tpexpr)) {
	case MSPH_TREE_CONJ:
		ret = msph_sphophi_foreach_pre_conj(f, sc,
		    (struct msph_tree_conj *)tpexpr);
		break;
	case MSPH_TREE_DISJ:
		ret = msph_sphophi_foreach_pre_disj(f, sc,
		    (struct msph_tree_disj *)tpexpr);
		break;
	case MSPH_TREE_IMPL:
		ret = msph_sphophi_foreach_pre_impl(f, sc,
		    (struct msph_tree_impl *)tpexpr);
		break;
	case MSPH_TREE_ARROW:
		ret = msph_sphophi_foreach_pre_arrow(f, sc,
		    (struct msph_tree_arrow *)tpexpr);
		break;
	case MSPH_TREE_TPNAME:
		ret = msph_sphophi_foreach_pre_tpname(f, sc,
		    (struct msph_tree_tpname *)tpexpr);
		break;
	case MSPH_TREE_TPAPPL:
		ret = msph_sphophi_foreach_pre_tpappl(f, sc,
		    (struct msph_tree_tpappl *)tpexpr);
		break;
	case MSPH_TREE_TRAIT:
		ret = msph_sphophi_foreach_pre_trait(f, sc,
		    (struct msph_tree_trait *)tpexpr);
		break;
	case MSPH_TREE_BOX:
		ret = msph_sphophi_foreach_pre_box(f, sc,
		    (struct msph_tree_box *)tpexpr);
		break;
	case MSPH_TREE_FORALL:
		ret = msph_sphophi_foreach_pre_forall(f, sc,
		    (struct msph_tree_forall *)tpexpr);
		break;
	case MSPH_TREE_PAREN:
		ret = msph_sphophi_foreach_pre_paren(f, sc,
		    (struct msph_tree_paren *)tpexpr);
		break;
	case MSPH_TREE_TRUE:
		ret = msph_sphophi_foreach_pre_true(f, sc,
		    (struct msph_tree_true *)tpexpr);
		break;
	case MSPH_TREE_FALSE:
		ret = msph_sphophi_foreach_pre_false(f, sc,
		    (struct msph_tree_false *)tpexpr);
		break;
	default:
		SPHO_RIP("default-reach");
		break;
	}

	return (ret);
}

static int
msph_sphophi_foreach_pre_conj(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_conj *conj)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(conj != NULL);

	if ((ret = f(sc, TREE(conj))) != 0)
		return (ret);

	if ((ret = msph_sphophi_foreach_pre_tpexpr(f, sc, conj->ltp)) != 0)
		return (ret);

	return (msph_sphophi_foreach_pre_tpexpr(f, sc, conj->rtp));
}

static int
msph_sphophi_foreach_pre_disj(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_disj *disj)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(disj != NULL);

	if ((ret = f(sc, TREE(disj))) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_pre_tpexpr(f, sc, disj->ltp)) != 0)
		return (ret);
	return (msph_sphophi_foreach_pre_tpexpr(f, sc, disj->rtp));
}

static int
msph_sphophi_foreach_pre_impl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_impl *impl)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(impl != NULL);

	if ((ret = f(sc, TREE(impl))) != 0)
		return (ret);

	if ((ret = msph_sphophi_foreach_pre_tpexpr(f, sc, impl->ltp)) != 0)
		return (ret);
	return (msph_sphophi_foreach_pre_tpexpr(f, sc, impl->rtp));
}

static int
msph_sphophi_foreach_pre_arrow(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_arrow *arrow)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(arrow != NULL);

	if ((ret = f(sc, TREE(arrow))) != 0)
		return (ret);

	if ((ret = msph_sphophi_foreach_pre_tpexpr(f, sc, arrow->ltp)) != 0)
		return (ret);
	return (msph_sphophi_foreach_pre_tpexpr(f, sc, arrow->rtp));
}

static int
msph_sphophi_foreach_pre_tpname(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_tpname *name)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(name != NULL);

	if ((ret = f(sc, TREE(name))) != 0)
		return (ret);

	return (msph_sphophi_foreach_pre_nameref(f, sc, name->name));
}

static int
msph_sphophi_foreach_pre_tpappl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_tpappl *appl)
{
	int ret;
	struct msph_tree_tpexpr *tp;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(appl != NULL);

	if ((ret = f(sc, TREE(appl))) != 0)
		return (ret);

	ret = 0;
	STAILQ_FOREACH(tp, &appl->args, entries) {
		if ((ret = msph_sphophi_foreach_pre_tpexpr(f, sc, tp)) != 0)
			return (ret);
	}

	return (ret);
}

static int
msph_sphophi_foreach_pre_trait(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_trait *trait)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(trait != NULL);

	if ((ret = f(sc, TREE(trait))) != 0)
		return (ret);

	sc = SCOPE(trait);

	return (msph_sphophi_foreach_pre_body(f, sc, trait->body));
}

static int
msph_sphophi_foreach_pre_box(msph_tree_sphophi_f f,struct spho_scope *sc,
    struct msph_tree_box *box)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(box != NULL);

	if ((ret = f(sc, TREE(box))) != 0)
		return (ret);

	return (msph_sphophi_foreach_pre_tpexpr(f, sc, box->inner));
}

static int
msph_sphophi_foreach_pre_forall(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_forall *forall)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(forall != NULL);

	if ((ret = f(sc, TREE(forall))) != 0)
		return (ret);

	sc = SCOPE(forall);

	return (msph_sphophi_foreach_pre_tpexpr(f, sc, forall->inner));
}

static int
msph_sphophi_foreach_pre_paren(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_paren *paren)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(paren != NULL);

	if ((ret = f(sc, TREE(paren))) != 0)
		return (ret);

	return (msph_sphophi_foreach_pre_tpexpr(f, sc, paren->inner));
}

static int
msph_sphophi_foreach_pre_true(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_true *tru)
{
	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(tru != NULL);

	return (f(sc, TREE(tru)));
}

static int
msph_sphophi_foreach_pre_false(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_false *fls)
{
	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(fls != NULL);

	return (f(sc, TREE(fls)));
}


/* foreach post: bottom up foreach */
static int
msph_sphophi_foreach_post(msph_tree_sphophi_f f, struct msph_tree_root *root)
{
	int ret;
	struct spho_scope *sc;
	struct msph_tree_unit *unit;

	SPHO_PRECOND(f != NULL);
	SPHO_PRECOND(root != NULL);
	SPHO_PRECOND(root->ctx != NULL);
	SPHO_PRECOND(root->ctx->spho != NULL);

	sc = spho_scope_global(root->ctx->spho);

	ret = 0;
	STAILQ_FOREACH(unit, &root->head, entries) {
		if ((ret = msph_sphophi_foreach_post_unit(f, sc, unit)) != 0)
			goto err;
	}

	if ((ret = f(sc, TREE(root))) != 0)
		goto err;

	return (ret);
err:
	MSPH_ERR(root->ctx, MSPH_ERR_SPHOPHI);
	return (ret);
}

static int
msph_sphophi_foreach_post_unit(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_unit *unit)
{
	int ret;
	struct spho_scope *inner_sc;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(unit != NULL);

	inner_sc = SCOPE(unit);

	if ((ret = msph_sphophi_foreach_post_body(f, inner_sc, unit->body))
	    != 0) {
		return (ret);
	}

	return (f(sc, TREE(unit)));
}

static int
msph_sphophi_foreach_post_body(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_body *body)
{
	int ret;
	struct msph_tree_dir *dir;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(body != NULL);

	ret = 0;
	STAILQ_FOREACH(dir, &body->head, entries) {
		if ((ret = msph_sphophi_foreach_post_dir(f, sc, dir)) != 0)
			return (ret);
	}

	return (f(sc, TREE(body)));
}

static int
msph_sphophi_foreach_post_dir(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_dir *dir)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(dir != NULL);

	ret = -1;
	switch (TREE_ID(dir)) {
	case MSPH_TREE_TPDEF:
		ret = msph_sphophi_foreach_post_tpdef(f, sc,
		    (struct msph_tree_tpdef *)dir);
		break;
	case MSPH_TREE_MEMBDECL:
		ret = msph_sphophi_foreach_post_membdecl(f, sc,
		    (struct msph_tree_membdecl *)dir);
		break;
	case MSPH_TREE_NOMINDECL:
		ret = msph_sphophi_foreach_post_nomindecl(f, sc,
		    (struct msph_tree_nomindecl *)dir);
		break;
	case MSPH_TREE_ASSERT:
		ret = msph_sphophi_foreach_post_assert(f, sc,
		    (struct msph_tree_assert *)dir);
		break;
	default:
		SPHO_ASSERT(0);
		break;
	}

	return (ret);
}

static int
msph_sphophi_foreach_post_tpdef(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_tpdef *tpdef)
{
	int ret;
	struct spho_scope *sub_sc;
	struct msph_tree_namedecl *name;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(tpdef != NULL);


	if ((ret = msph_sphophi_foreach_post_namedecl(f, sc, tpdef->name))
	    != 0) {
		return (ret);
	}

	sub_sc = HAS_SCOPE(tpdef) ? SCOPE(tpdef) : sc;

	STAILQ_FOREACH(name, &tpdef->params, entries) {
		if ((ret = msph_sphophi_foreach_post_namedecl(f, sub_sc, name))
		    != 0) {
			return (ret);
		}
	}

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sub_sc, tpdef->tp))
	    != 0)
		return (ret);

	return (f(sc, TREE(tpdef)));
}

static int
msph_sphophi_foreach_post_nomindecl(msph_tree_sphophi_f f,
    struct spho_scope *sc, struct msph_tree_nomindecl *nomd)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(nomd != NULL);

	if ((ret = msph_sphophi_foreach_post_namedecl(f, sc, nomd->name)) != 0)
		return (ret);

	return (f(sc, TREE(nomd)));
}

static int
msph_sphophi_foreach_post_membdecl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_membdecl *membd)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(membd != NULL);

	if ((ret = msph_sphophi_foreach_post_namedecl(f, sc, membd->name)) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, membd->tp)) != 0)
		return (ret);

	return (f(sc, TREE(membd)));
}

static int
msph_sphophi_foreach_post_assert(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_assert *ass)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(ass != NULL);

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, ass->ltp)) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, ass->rtp)) != 0)
		return (ret);

	return (f(sc, TREE(ass)));
}

static int
msph_sphophi_foreach_post_namedecl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_namedecl *name)
{
	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(name != NULL);

	return (f(sc, TREE(name)));
}

static int
msph_sphophi_foreach_post_nameref(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_nameref *name)
{
	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(name != NULL);

	return (f(sc, TREE(name)));
}

static int
msph_sphophi_foreach_post_tpexpr(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_tpexpr *tpexpr)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(tpexpr != NULL);

	ret = -1;

	switch (TREE_ID(tpexpr)) {
	case MSPH_TREE_CONJ:
		ret = msph_sphophi_foreach_post_conj(f, sc,
		    (struct msph_tree_conj *)tpexpr);
		break;
	case MSPH_TREE_DISJ:
		ret = msph_sphophi_foreach_post_disj(f, sc,
		    (struct msph_tree_disj *)tpexpr);
		break;
	case MSPH_TREE_IMPL:
		ret = msph_sphophi_foreach_post_impl(f, sc,
		    (struct msph_tree_impl *)tpexpr);
		break;
	case MSPH_TREE_ARROW:
		ret = msph_sphophi_foreach_post_arrow(f, sc,
		    (struct msph_tree_arrow *)tpexpr);
		break;
	case MSPH_TREE_TPNAME:
		ret = msph_sphophi_foreach_post_tpname(f, sc,
		    (struct msph_tree_tpname *)tpexpr);
		break;
	case MSPH_TREE_TPAPPL:
		ret = msph_sphophi_foreach_post_tpappl(f, sc,
		    (struct msph_tree_tpappl *)tpexpr);
		break;
	case MSPH_TREE_TRAIT:
		ret = msph_sphophi_foreach_post_trait(f, sc,
		    (struct msph_tree_trait *)tpexpr);
		break;
	case MSPH_TREE_BOX:
		ret = msph_sphophi_foreach_post_box(f, sc,
		    (struct msph_tree_box *)tpexpr);
		break;
	case MSPH_TREE_FORALL:
		ret = msph_sphophi_foreach_post_forall(f, sc,
		    (struct msph_tree_forall *)tpexpr);
		break;
	case MSPH_TREE_PAREN:
		ret = msph_sphophi_foreach_post_paren(f, sc,
		    (struct msph_tree_paren *)tpexpr);
		break;
	case MSPH_TREE_TRUE:
		ret = msph_sphophi_foreach_post_true(f, sc,
		    (struct msph_tree_true *)tpexpr);
		break;
	case MSPH_TREE_FALSE:
		ret = msph_sphophi_foreach_post_false(f, sc,
		    (struct msph_tree_false *)tpexpr);
		break;
	default:
		SPHO_ASSERT(0);
		break;
	}

	return (ret);
}

static int
msph_sphophi_foreach_post_conj(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_conj *conj)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(conj != NULL);

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, conj->ltp)) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, conj->rtp)) != 0)
		return (ret);

	return (f(sc, TREE(conj)));

}

static int
msph_sphophi_foreach_post_disj(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_disj *disj)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(disj != NULL);

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, disj->ltp)) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, disj->rtp)) != 0)
		return (ret);

	return (f(sc, TREE(disj)));
}

static int
msph_sphophi_foreach_post_impl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_impl *impl)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(impl != NULL);

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, impl->ltp)) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, impl->rtp)) != 0)
		return (ret);

	return (f(sc, TREE(impl)));
}

static int
msph_sphophi_foreach_post_arrow(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_arrow *arrow)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(arrow != NULL);

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, arrow->ltp)) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, arrow->rtp)) != 0)
		return (ret);

	return (f(sc, TREE(arrow)));
}

static int
msph_sphophi_foreach_post_tpname(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_tpname *name)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(name != NULL);

	if ((ret = msph_sphophi_foreach_post_nameref(f, sc, name->name)) != 0)
		return (ret);

	return (f(sc, TREE(name)));
}

static int
msph_sphophi_foreach_post_tpappl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_tpappl *appl)
{
	int ret;
	struct msph_tree_tpexpr *tp;

	ret = 0;
	STAILQ_FOREACH(tp, &appl->args, entries) {
		if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, tp)) != 0)
			return (ret);
	}

	return (f(sc, TREE(appl)));
}

static int
msph_sphophi_foreach_post_trait(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_trait *trait)
{
	int ret;
	struct spho_scope *inner_sc;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(trait != NULL);

	inner_sc = SCOPE(trait);

	if ((ret = msph_sphophi_foreach_post_body(f, inner_sc, trait->body))
	    != 0) {
		return (ret);
	}

	return (f(sc, TREE(trait)));
}

static int
msph_sphophi_foreach_post_box(msph_tree_sphophi_f f,struct spho_scope *sc,
    struct msph_tree_box *box)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(box != NULL);

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, box->inner)) != 0)
		return (ret);

	return (f(sc, TREE(box)));
}

static int
msph_sphophi_foreach_post_forall(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_forall *forall)
{
	int ret;
	struct spho_scope *inner_sc;
	
	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(forall != NULL);

	inner_sc = SCOPE(forall);

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, inner_sc, forall->inner))
	    != 0)
		return (ret);

	return (f(sc, TREE(forall)));

}

static int
msph_sphophi_foreach_post_paren(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_paren *paren)
{
	int ret;

	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(paren != NULL);

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, paren->inner)) != 0)
		return (ret);

	return (f(sc, TREE(paren)));

}

static int
msph_sphophi_foreach_post_true(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_true *tru)
{
	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(tru != NULL);

	return (f(sc, TREE(tru)));
}

static int
msph_sphophi_foreach_post_false(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_false *fls)
{
	SPHO_PRECOND(sc != NULL);
	SPHO_PRECOND(fls != NULL);

	return (f(sc, TREE(fls)));
}
