/*
 * Subtyping
 */
#include <stdlib.h>

#include "spho/bind.h"
#include "spho/bind.h"
#include "spho/spho.h"

#include "spho/subt.h"

struct bound_tp {
	struct spho_bind		*bind;
	const struct spho_tp		*tp;
};

struct seqnt {
	struct spho_ctx		*ctx;

	size_t			sz;
	size_t			filled;
	struct bound_tp		*elms;
};

struct seqnt_tstl {
	struct seqnt		left;
	struct seqnt		right;
};


static struct bound_tp *spho_bound_tp(const struct spho_tp *);


/* read scopes and create bound type */
struct bound_tp *
spho_bound_tp(const struct spho_tp *tp)
{
	struct spho_scope	*sc;
	struct spho_bind	*bind, *more_bind;

	struct bound_tp	*ret;

	SPHO_PRECOND(tp != NULL);
	SPHO_PRECOND(tp->sc != NULL);
	SPHO_PRECOND(tp->sc->bind_loc != NULL);

	sc = tp->sc;
	bind = NULL;

	if ((ret = calloc(1, sizeof(*ret))) == NULL) {
		SPHO_ERR(sc->ctx, SPHO_ERR_SYS);
		return (NULL);
	}
	ret->tp = tp;

	
	if ((ret->bind = calloc(1, sizeof(*ret->bind))) == NULL) {
		SPHO_ERR(sc->ctx, SPHO_ERR_SYS);
		goto err;
	}

	/* initialize parent and duplicate scope-local bindings */
	ret->bind->parent = NULL;
	if ((ret->bind->loc = spho_prebind_dupl(sc->bind_loc)) == NULL) {
		goto err;
	}

	/* iterate up through parent scopes and gather their local bindings */
	bind = ret->bind;
	while ((sc = sc->parent) != NULL) {
		if ((more_bind = calloc(1, sizeof(*more_bind))) == NULL) {
			SPHO_ERR(sc->ctx, SPHO_ERR_SYS);
			goto err;
		}
		more_bind->parent = NULL;
		if ((more_bind->loc = spho_prebind_dupl(sc->bind_loc))
		    == NULL) {
			goto err;
		}
		bind->parent = more_bind;
		bind = bind->parent;
	}

	return (ret);

err:
	while (ret->bind != NULL) {
		bind = ret->bind;
		ret->bind = ret->bind->parent;

		spho_prebind_free(bind->loc);
		free(bind);
	}

	return (NULL);
}

static struct seqnt *
seqnt_alloc(struct spho_ctx *ctx, size_t initsz)
{
	struct seqnt *seq;
	if (initsz == 0)
		initsz = 32; /* just feels good */

	if ((seq = calloc(1, sizeof(*seq))) == NULL) {
		SPHO_ERR(ctx, SPHO_ERR_SYS);
		return (NULL);
	}
	seq->ctx = ctx;

	if ((seq->elms = calloc(initsz, sizeof(*seq->elms))) == NULL) {
		SPHO_ERR(seq->ctx, SPHO_ERR_SYS);
		goto err;
	}

	seq->sz = initsz;
	seq->filled = 0;

	return (seq);
err:
	if (seq != NULL) {
		if (seq->elms != NULL)
			free(seq->elms);
		free(seq);
	}

	return (NULL);
}

static int
seqnt_makebig(struct seqnt *seq)
{
	size_t		new_sz;
	struct bound_tp *new_elms;

	SPHO_PRECOND(seq != NULL);
	SPHO_PRECOND(seq->sz > 0);
	SPHO_PRECOND(seq->sz <= (SIZE_MAX >> 1));
	SPHO_PRECOND(seq->elms != NULL);

	new_sz = seq->sz << 1;

	if ((new_elms = realloc(seq->elms, new_sz)) == NULL) {
		SPHO_ERR(seq->ctx, SPHO_ERR_SYS);
		return (-1);
	}

	seq->sz = new_sz;
	seq->elms = new_elms;

	return (0);
}

static int
seqnt_insert_from_ptr(struct seqnt *seq, struct bound_tp *elm)
{
	SPHO_PRECOND(seq->filled <= seq->sz);

	if (seq->filled >= seq->sz && (seqnt_makebig(seq) == -1))
		return (-1);

	seq->elms[seq->filled++] = *elm;

	SPHO_POSTCOND(seq->filled <= seq->sz);

	return (0);
}

static int
seqnt_insert_bind_tp(struct seqnt *seq, struct spho_bind *elm_bind,
    struct spho_tp *elm_tp)
{
	SPHO_PRECOND(seq->filled <= seq->sz);

	if (seq->filled >= seq->sz && (seqnt_makebig(seq) == -1))
		return (-1);

	seq->elms[seq->filled++] = (struct bound_tp){ elm_bind, elm_tp };

	SPHO_POSTCOND(seq->filled <= seq->sz);

	return (0);
}


int
spho_is_subt(struct spho_tp *tp_left, struct spho_tp *tp_right)
{
	struct bound_tp *left, *right;

	if ((left = spho_bound_tp(tp_left)) == NULL)
		return (-1);
	if ((right = spho_bound_tp(tp_right)) == NULL)
		return (-1);

	// TODO check well-definedness (e.g. no undefs)

	// XXX skip boxing step
	
	// TODO create seqnt and seqnt_tstl
	// TODO run decider

	return (-1);
}

